/* 
 * Copyright 2004 The Apache Software Foundation
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * Edited by Bjrn Harald Olsen 2006
 */
using System;
using Analyzer = Lucene.Net.Analysis.Analyzer;
using LowerCaseFilter = Lucene.Net.Analysis.LowerCaseFilter;
using StopFilter = Lucene.Net.Analysis.StopFilter;
using TokenStream = Lucene.Net.Analysis.TokenStream;
using StandardFilter = Lucene.Net.Analysis.Standard.StandardFilter;
using StandardTokenizer = Lucene.Net.Analysis.Standard.StandardTokenizer;
namespace NorwegianAnalyzer
{
	
	/// <summary> Analyzer for Norwegian language. Supports an external list of stopwords (words that
	/// will not be indexed at all) 
	/// A default set of stopwords is used unless an alternative list is specified.
	/// 
	/// </summary>
	/// <author>  Bjrn Harald Olsen
	/// </author>
	/// <version>  $Id: NorwegianAnalyzer.cs
	/// </version>
	public class NorwegianAnalyzer : Analyzer
	{
		/// <summary> List of typical norwegian stopwords.</summary>
		private System.String[] NORWEGIAN_STOP_WORDS = new System.String[]
			{
				"i","og"
			};
		
		/// <summary> Contains the stopwords used with the StopFilter.</summary>
		private System.Collections.Hashtable stopSet = new System.Collections.Hashtable();
				
		/// <summary> Builds an analyzer.</summary>
		public NorwegianAnalyzer()
		{
			stopSet = StopFilter.MakeStopSet(NORWEGIAN_STOP_WORDS);
		}
		
		/// <summary> Builds an analyzer with the given stop words.</summary>
		public NorwegianAnalyzer(System.String[] stopwords)
		{
			stopSet = StopFilter.MakeStopSet(stopwords);
		}
		
		/// <summary> Builds an analyzer with the given stop words.</summary>
		public NorwegianAnalyzer(System.Collections.Hashtable stopwords)
		{
			stopSet = new System.Collections.Hashtable(new System.Collections.Hashtable(stopwords));
		}
		
		/// <summary> Builds an analyzer with the given stop words.</summary>
		public NorwegianAnalyzer(System.IO.FileInfo stopwords)
		{
			stopSet = WordlistLoader.GetWordSet(stopwords);
		}
		
		/// <summary> Creates a TokenStream which tokenizes all the text in the provided Reader.
		/// 
		/// </summary>
		/// <returns> A TokenStream build from a StandardTokenizer filtered with
		/// StandardFilter, LowerCaseFilter, StopFilter, GermanStemFilter
		/// </returns>
		public override TokenStream TokenStream(System.String fieldName, System.IO.TextReader reader)
		{
			TokenStream result = new StandardTokenizer(reader);
			result = new StandardFilter(result);
			result = new LowerCaseFilter(result);
			result = new StopFilter(result, stopSet);
			//result = new NoiseFilter(result);			
			result = new NorwegianStemFilter(result);			
			return result;
		}
	}
}