/* 
 * Copyright 2004 The Apache Software Foundation
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * Edited by Bjrn Harald Olsen 2006
 */
using System;
namespace NorwegianAnalyzer
{
	/// <summary> A stemmer for norwegian words. The algorithm is based on... 
	/// 
	/// </summary>
	/// <author>     Bjrn Harald Olsen
	/// </author>
	/// <version>    $Id: NorwegianStemmer.cs
	/// </version>
	public class NorwegianStemmer
	{		
		private char[] b;
		private int i, j, k, k0;
		private bool dirty = false;
		private bool stemmed = false;
        private int region1;
		private const int INC = 50; /* unit of size whereby b is increased */
		private const int EXTRA = 1;
		
		public NorwegianStemmer()
		{
			b = new char[INC];
			i = 0;
		}
		
		/// <summary> reset() resets the stemmer so it can stem another word.  If you invoke
		/// the stemmer by calling add(char) and then stem(), you must call reset()
		/// before starting another word.
		/// </summary>
		public virtual void  Reset()
		{
			i = 0; dirty = false; stemmed=false;
            region1 = -1;
		}
		
		/// <summary> Add a character to the word being stemmed.  When you are finished
		/// adding characters, you can call stem(void) to process the word.
		/// </summary>
		public virtual void  Add(char ch)
		{
			if (b.Length <= i + EXTRA)
			{
				char[] new_b = new char[b.Length + INC];
				for (int c = 0; c < b.Length; c++)
					new_b[c] = b[c];
				b = new_b;
			}
			b[i++] = ch;
		}
		
		/// <summary> After a word has been stemmed, it can be retrieved by toString(),
		/// or a reference to the internal buffer can be retrieved by getResultBuffer
		/// and getResultLength (which is generally more efficient.)
		/// </summary>
		public override System.String ToString()
		{
			return new System.String(b, 0, i);
		}
		
		/// <summary> Returns the length of the word resulting from the stemming process.</summary>
		public virtual int GetResultLength()
		{
			return i;
		}
		
		/// <summary> Returns a reference to a character buffer containing the results of
		/// the stemming process.  You also need to consult getResultLength()
		/// to determine the length of the result.
		/// </summary>
		public virtual char[] GetResultBuffer()
		{
			return b;
		}
		
		/* cons(i) is true <=> b[i] is a consonant. */
		
		private bool Cons(int i)
		{
			switch (b[i])
			{
				
				case 'a': 
				case 'e': 
				case 'i': 
				case 'o': 
				case 'u':
				case 'y':
				case '':
				case '':
				case '':
					return false;				
				
				default: 
					return true;
				
			}
		}
		
		/* m() measures the number of consonant sequences between k0 and j. if c is
		a consonant sequence and v a vowel sequence, and <..> indicates arbitrary
		presence,
		
		<c><v>       gives 0
		<c>vc<v>     gives 1
		<c>vcvc<v>   gives 2
		<c>vcvcvc<v> gives 3
		....
		*/
		
		private int M()
		{
			int n = 0;
			int i = k0;
			while (true)
			{
				if (i > j)
					return n;
				if (!Cons(i))
					break;
				i++;
			}
			i++;
			while (true)
			{
				while (true)
				{
					if (i > j)
						return n;
					if (Cons(i))
						break;
					i++;
				}
				i++;
				n++;
				while (true)
				{
					if (i > j)
						return n;
					if (!Cons(i))
						break;
					i++;
				}
				i++;
			}
		}
		
		/* vowelinstem() is true <=> k0,...j contains a vowel */
		
		private bool vowelinstem()
		{
			int i;
			for (i = k0; i <= j; i++)
				if (!Cons(i))
					return true;
			return false;
		}
		
		/* doublec(j) is true <=> j,(j-1) contain a double consonant. */
		
		private bool Doublec(int j)
		{
			if (j < k0 + 1)
				return false;
			if (b[j] != b[j - 1])
				return false;
			return Cons(j);
		}
		
		/* cvc(i) is true <=> i-2,i-1,i has the form consonant - vowel - consonant
		and also if the second c is not w,x or y. this is used when trying to
		restore an e at the end of a short word. e.g.
		
		cav(e), lov(e), hop(e), crim(e), but
		snow, box, tray.
		
		*/
		
		private bool Cvc(int i)
		{
			if (i < k0 + 2 || !Cons(i) || Cons(i - 1) || !Cons(i - 2))
				return false;
			else
			{
				int ch = b[i];
				if (ch == 'w' || ch == 'x' || ch == 'y')
					return false;
			}
			return true;
		}
		
		private bool Ends(System.String s)
		{            
            if (region1 < 0)
            {         
                return false;
            }

			int l = s.Length;
			int o = k - l + 1;
            if (o < k0)
            {                
                return false;
            }
			for (int i = 0; i < l; i++)
                if (b[o + i] != s[i])
                {
                    return false;
                }
            if (k - l < region1-1)
            {
                return false;
            }            

			j = k - l;
            return true;
		}

        private bool Ends_gen(System.String s)
        {
            if (region1 < 0)
            {             
                return false;
            }
            int l = s.Length;
            int o = k - l + 1;
            if (o < k0)
            {         
                return false;
            }
            for (int i = 0; i < l; i++)
                if (b[o + i] != s[i])
                {
                    return false;
                }            
            j = k - l;
            return true;
        }
		
		/* setto(s) sets (j+1),...k to the characters in the string s, readjusting
		k. */
		
		internal virtual void  Setto(System.String s)
		{
			int l = s.Length;
			int o = j + 1;
			for (int i = 0; i < l; i++)
				b[o + i] = s[i];
			k = j + l;
			dirty = true;
		}
		
		/* r(s) is used further down. */
		
		internal virtual void  R(System.String s)
		{
			if (M() > 0)
				Setto(s);
		}
		
		/* step1() gets rid of plurals and -ed or -ing. e.g.		
		*/

		private bool abridge_extractor()
		{				
			return false;
			/*
			if(Ends("ene")|| Ends("er")) k=j+1;
			else return false;
			
			string[] variants = {"kle", "tre", "sle", "mle", "ple", "mne"};	
			string[] var_set =  {"kel", "ter", "sel", "mel", "pel", "men"};
			string var="";

			for (int _i=0; _i < variants.Length; _i++)
			{
				if(Ends(variants[_i]))
				{
					if( !Cons(j-1) )
					{			
						Setto(var_set[_i]);
						return true;
					}
					else
					{		
						var=var_set[_i].Insert(0,var_set[_i].Substring(0,1));
						Setto(var);
						return true;
					}
				}
			}
			k=i-1;
			return false;
			*/
		}
		
		private void  step1()
		{            
            if (Ends("a"))
            {             
                if (j >= 2) k--;
                return;
            }            
            else if (b[k] == 'e')
            {
                if (Ends("erte"))
                {
                    k -= 2; /* -erte -> -er */
                    return;
                }
                else
                {             
                    string[] ends_e = { "hetene", "ande", "ende", "ede", "ane", "ene", "e" };
                    foreach (string e in ends_e) 
                        if (Ends(e)) 
                        {
                            if (j >= 2)
                            {
                                stemmed = true;
                                break;
                            }
                        }

                }
            }
            else if (b[k] == 'n')
            {                
                string[] ends_n = { "heten", "en" };
                foreach (string e in ends_n)
                {
                    if (Ends(e))
                    {
                        if (j >= 2)
                        {
                            stemmed = true;
                            break;
                        }
                    }
                }
            }
            else if (b[k] == 'r')
            {
                if (abridge_extractor() && Ends("er")) stemmed = true;
                else
                {
                    string[] ends_r = { "heter", "er", "ar" };
                    foreach (string e in ends_r) 
                        if (Ends(e))
                        {
                            if (j >= 2)
                            {
                                stemmed = true;
                                break;
                            }
                        }
                }
            }
            else if (b[k] == 's')
            {
                string[] ends_s = { "hetenes", "hetens", "endes", "edes", "enes", "ens", "ers", "ets", "as", "es" };
                foreach (string e in ends_s) 
                    if (Ends(e)) 
                    {
                        if (j >= 2)
                        {
                            stemmed = true;
                            break;
                        }
                    }

                if (!stemmed)
                {
                    if (Ends_gen("bs") ||
                        Ends_gen("cs") ||
                        Ends_gen("ds") ||
                        Ends_gen("fs") ||
                        Ends_gen("gs") ||
                        Ends_gen("hs") ||
                        Ends_gen("js") ||
                        Ends_gen("ls") ||
                        Ends_gen("ms") ||
                        Ends_gen("ns") ||
                        Ends_gen("os") ||
                        Ends_gen("ps") ||
                        Ends_gen("rs") ||
                        Ends_gen("ts") ||
                        Ends_gen("vs") ||
                        Ends_gen("ys") ||
                        Ends_gen("zs"))
                    {                        
                        if(j>0 && vowelinstem() && region1>0 && region1<i) k--;
                        return;
                    }
                    if (b[k - 1] == 'k' && Cons(k - 2)) 
                    {
                        k--;
                        return;
                    }
                }
            }
            else if (b[k] == 't')
            {
                if (Ends("ert"))
                {
                    k--; /* -ert -> -er */
                    return;
                }
                else
                {                   
                    string[] ends_t = { "het", "ast", "et" };
                    foreach (string e in ends_t) 
                        if (Ends(e)) 
                        {
                            if (j >= 2)
                            {
                                stemmed = true;
                                break;
                            }
                        }
                }
            }

            if (stemmed && i>3)
            {
                //Console.WriteLine("setter ny k: " + j);
                k = j;
            }            
		}
		
		/* step2() turns terminal y to i when there is another vowel in the stem. */
		
		private void  step2()
		{
            if (Ends("dt") || Ends("vt"))
                if (j >= 2) k--;            
		}

		private void step3()
		{
            /*
            if ((
                Ends("eleg") ||    // variant of -elig- 
                Ends("leg") ||     // variant of -lig- 
                Ends("elig") ||
                Ends("lig") ||
                Ends("eig") ||
                Ends("ig") ||
                Ends("elov") ||
                Ends("hetslov") || // Rare. -elseslov- is possible, but very rare 
                Ends("slov") ||
                Ends("lov") ||
                Ends("els")))
            {
                if (j >= 2 && (region1 > 0 && region1 < i)) k = j;
            }*/
             
            string[] ends_s = {"hetslov","elig","elov","slov","eleg","leg","lig","eig","lov","els","ig"};
            foreach (string e in ends_s)
                if (Ends(e))
                {
                    if (j >= 2 && region1 > 0 && region1 < i)
                    {
                        k = j;
                        break;
                    }
                }
            
		}
		
		/// <summary> Stem a word provided as a String.  Returns the result as a String.</summary>
		public virtual System.String Stem(System.String s)
		{
			//Console.Write(s+" --> ");
			if (Stem(s.ToCharArray(), s.Length))
			{	
				//Console.WriteLine(ToString() /*+ " i:"+i+" j:"+j+" k:"+k+" k0:"+k0+" M: "+M()*/);
				return ToString();
			}
			else
			{				
				//Console.WriteLine(s);
				return s;
			}
		}
		
		/// <summary>Stem a word contained in a char[].  Returns true if the stemming process
		/// resulted in a word different from the input.  You can retrieve the
		/// result with getResultLength()/getResultBuffer() or toString().
		/// </summary>
		public virtual bool Stem(char[] word)
		{
			return Stem(word, word.Length);
		}
		
		/// <summary>Stem a word contained in a portion of a char[] array.  Returns
		/// true if the stemming process resulted in a word different from
		/// the input.  You can retrieve the result with
		/// getResultLength()/getResultBuffer() or toString().
		/// </summary>
		public virtual bool Stem(char[] wordBuffer, int offset, int wordLen)
		{
			Reset();
			if (b.Length < wordLen)
			{
				char[] new_b = new char[wordLen + EXTRA];
				b = new_b;
			}
            int foundVowel = -1;
            
            for (int j = 0; j < wordLen; j++)
            {
                b[j] = wordBuffer[offset + j];
                if (region1 < 0)
                {
                    if (Cons(j) && foundVowel==(j-1) && foundVowel>=0 ) region1 = j+1;
                    if(!Cons(j))
                        foundVowel = j;
                }
            }
			i = wordLen;
			return Stem(0);
		}
		
		/// <summary>Stem a word contained in a leading portion of a char[] array.
		/// Returns true if the stemming process resulted in a word different
		/// from the input.  You can retrieve the result with
		/// getResultLength()/getResultBuffer() or toString().
		/// </summary>
		public virtual bool Stem(char[] word, int wordLen)
		{
			return Stem(word, 0, wordLen);
		}
		
		/// <summary>Stem the word placed into the Stemmer buffer through calls to add().
		/// Returns true if the stemming process resulted in a word different
		/// from the input.  You can retrieve the result with
		/// getResultLength()/getResultBuffer() or toString().
		/// </summary>
		public virtual bool Stem()
		{
			return Stem(0);
		}
		
		public virtual bool Stem(int i0)
		{
            //Console.WriteLine("\n\n\n");
			k = i - 1;
			k0 = i0;
			if ( k > k0 + 1 )
			{				
				step1(); step2(); step3();				
			}

            //Console.WriteLine("RESULTAT k: " + k);
			
			// Also, a word is considered dirty if we lopped off letters
			// Thanks to Ifigenia Vairelles for pointing this out.
			if (i != k + 1)
				dirty = true;			
			i = k + 1;
			return dirty;
		}		
		
		/// <summary> Checks if a term could be stemmed.
		/// 
		/// </summary>
		/// <returns>  true if, and only if, the given term consists in letters.
		/// </returns>
		private bool IsStemmable(System.String term)
		{
			for (int c = 0; c < term.Length; c++)
			{
				if (!System.Char.IsLetter(term[c]))
					return false;
			}
			return true;
		}		
	}
}
