/*
 * Author Bjrn Harald Olsen 2006
 * 
 * 
 */
using System;
using System.Diagnostics;
using System.IO;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Documents;
using Lucene.Net.Index;
using Lucene.Net.QueryParsers;
using Lucene.Net.Search;
using Analyzer = Lucene.Net.Analysis.Analyzer;

namespace TrelsEval
{
	public class TrelsEval
	{		
		private string pathIndex;
		private string pathDoc;
		private string pathTrels;
        StreamWriter sw;
		private IndexWriter indexWriter;
		private string[] patterns = {"*.doc", "*.xls", "*.ppt", "*.htm", "*.html", "*.txt", "*.pdf"};		
		
		private IndexSearcher searcher = null;
        private Analyzer analyzer;
		private ArrayList trelsSet;

		private double BETA = 1.0;

		// statistics
		private long bytesTotal = 0;
		private int countTotal = 0;
		private int countSkipped = 0;

		public TrelsEval(string pathDoc, string pathIndex, string pathTrels, Analyzer analyzer, string outfile)
		{
			this.pathIndex = pathIndex;
			this.pathDoc = pathDoc;			
			this.pathTrels = pathTrels;
            sw = new StreamWriter(outfile);
			this.trelsSet = new ArrayList();
            this.analyzer = analyzer;
            Console.Write("Rebuild index? (Warning: may take several hours)[y/N] ");
            if (Console.ReadLine() == "y")
            {
                Console.WriteLine("Bygger index");
                buildIndex();
            }
            loadTrels(new DirectoryInfo(pathTrels));                       
            search();
            //calcMeanWordLength(new DirectoryInfo(pathDoc));
		}

        private void calcMeanWordLength(DirectoryInfo directory)
        {
            int sum = 0;
            long ant = 0;
            string path;
            foreach (string pattern in patterns)
            {
                foreach (FileInfo fi in directory.GetFiles(pattern))
                {
                    path = fi.FullName;
                    string text = "";
                    if (pattern == "*.pdf")
                        text = AdobeParser.Parse(path);
                    else if (pattern == "*.txt")
                        text = parsePlain(path);
                    else
                        text = Parser.Parse(path);

                    MatchCollection docCount = Regex.Matches(text, "[a-,.] ");
                    sum += docCount.Count;                    
                    ant++;
                }
            }
            Console.WriteLine("mean words: "+(sum/ant));
        }

		private void loadTrels(DirectoryInfo trelsDir)
		{
			RelevanceSet buf;
			foreach (FileInfo fi in trelsDir.GetFiles("*.xml"))
			{
				// skip temporary office files
				if (fi.Name.StartsWith("~"))
					continue;

				try 
				{
					buf = TrelsXmlReader.Parse(fi);
					if(buf!=null) trelsSet.Add(buf);
				}
				catch (Exception e)
				{					
					Console.WriteLine(e.ToString());
				}
			}
		}

		private void buildIndex()
		{
			indexWriter = new IndexWriter(pathIndex, analyzer, true);

			bytesTotal = 0;
			countTotal = 0;
			countSkipped = 0;

			DirectoryInfo di = new DirectoryInfo(pathDoc);
			
			DateTime start = DateTime.Now;
			
			addFolder(di);

			string summary = String.Format("Done. Indexed {0} files ({1} bytes). Skipped {2} files.", countTotal, bytesTotal, countSkipped);
			summary += String.Format(" Took {0}", (DateTime.Now - start));
			Console.WriteLine(summary);			

			indexWriter.Optimize();
			indexWriter.Close();	
		}

		private void addFolder(DirectoryInfo directory)
		{
			// find all matching files
			foreach (string pattern in patterns)
			{
				foreach (FileInfo fi in directory.GetFiles(pattern))
				{
					// skip temporary office files
					if (fi.Name.StartsWith("~"))
						continue;

					try 
					{
                        addDocument(fi.FullName, pattern);                       

						// update statistics
						this.countTotal++;
						this.bytesTotal += fi.Length;

						// show added file
						//status(fi.FullName);
					}
					catch (Exception)
					{
						// parsing and indexing wasn't successful, skipping that file
						this.countSkipped++;
						Console.WriteLine("Skipped: " + fi.FullName);
					}
				}
			}

			// add subfolders
			foreach (DirectoryInfo di in directory.GetDirectories())
			{
				addFolder(di);
			}
		}

        private void addDocument(string path, string pattern)
        {
            string text = "";
            if (pattern == "*.pdf")
                text = AdobeParser.Parse(path);
            else if (pattern == "*.txt")
                text = parsePlain(path);
            else
                text = Parser.Parse(path);

            MatchCollection docCount = Regex.Matches(text, "[a-,.] ");
            int docSize = docCount.Count;
            
            Document doc = new Document();
			string filename = Path.GetFileName(path);
            
            if(docSize<5) doc.SetBoost(0.3f);
            else if (docSize < 10) doc.SetBoost(0.6f);
            else if (docSize < 15) doc.SetBoost(0.9f);

            doc.Add(Field.UnStored("text", text));
			doc.Add(Field.Keyword("path", path));
			doc.Add(Field.Text("title", filename));
         
			indexWriter.AddDocument(doc);
        }	

        private string parsePlain(string path)
        {
            string text = "";
            using (StreamReader sr = File.OpenText(path))
            {
                string s = "";
                while ((s = sr.ReadLine()) != null)
                {
                    text += s + " ";
                }
            }
            return text;
        }
		
		private bool checkIndex()
		{
			try 
			{
				searcher = new IndexSearcher(this.pathIndex);
				searcher.Close();
			}
			catch (IOException)
			{
				//status("The index doesn't exist or is damaged. Please rebuild it.", true);
				return false;
			}
			
			string msg = String.Format("Index is ready. It contains {0} documents.", searcher.MaxDoc());
            return true;
			//status(msg);
		}

		private void search()
		{
            ArrayList snitt3 = new ArrayList();
            ArrayList snitt4 = new ArrayList();
			DateTime start = DateTime.Now;
            
			try 
			{
				searcher = new IndexSearcher(this.pathIndex);
			}
			catch (IOException ex)
			{

				Console.WriteLine("The index doesn't exist or is damaged. Please rebuild the index.\r\n\r\nDetails:\r\n" + ex.Message);
				return;
			}			
			
			foreach (RelevanceSet trels in trelsSet)
			{   
                ///// OUT
				Console.WriteLine(trels.getQuery());             
                sw.WriteLine(trels.getQuery());
                ///// OUT

				Query query = QueryParser.Parse( trels.getQuery() , "text" , analyzer);
                //Query query = QueryParser.Parse("\""+trel.getQuery()+"\"", "text", analyzer);                
                Hits hits = searcher.Search(query);
								
				ArrayList  scores = new ArrayList();
                
                if (hits.Length() < 10) Console.WriteLine(">> obs obs, mindre enn 10 treff");
				for (int i = 0; i < Math.Min(10,hits.Length()); i++) 
				{					
					Document doc = hits.Doc(i);
                    
					double numOn=0;
				    double numOff=0;
                    					
                    string path = doc.Get("path");                    
                    FileInfo fi = new FileInfo(path);
                    string text = "";

                    if (fi.Extension == ".pdf")
                        text = AdobeParser.Parse(path);
                    else if (fi.Extension == ".txt") // no need to filter a plain text file
                    {
                        text = parsePlain(path);
                    }
                    else
                        text = Parser.Parse(path);

                    text = text.ToLower();

					foreach (string term in trels.getOnTopics())
					{
                        //MatchCollection matches = Regex.Matches(text, ont + "[,. ]");
                        MatchCollection matches = Regex.Matches(text, term);
                        numOn += matches.Count;                        
					}
					foreach (string term in trels.getOffTopics())
					{
                        //MatchCollection matches = Regex.Matches(text, off + "[,. ]");
                        MatchCollection matches = Regex.Matches(text, term);
                        numOff += matches.Count;                                         
					}
                    MatchCollection docCount = Regex.Matches(text, "[a-,.] ");
                    double docSize = (double)docCount.Count / 100.0;
					double tScore_basic = (numOn - ( BETA*numOff ));
                    ////// OUT
                    Console.WriteLine(path + " score: "+hits.Score(i)+" docsize: "+docSize);
                    sw.WriteLine(path + " score: " + hits.Score(i));
                    ////// OUT
					scores.Add(tScore_basic);					
				} 
				if(hits.Length()==0) Console.WriteLine("No hits");
				else
				{
                    double tsD3 = tScoreD3(scores);
                    double tsD4 = tScoreD4(scores);
                    snitt3.Add(tsD3);
                    snitt4.Add(tsD4);
                    ////// OUT
                    Console.WriteLine("d3: " + tsD3 + " d4: " + tsD4);
                    //sw.WriteLine("d3: " + tsD3 + " d4: " + tsD4);
                    sw.WriteLine(tsD4);
                    ////// OUT
				}				
			}
            //sw.WriteLine("snitt3: " + tScoreD4(snitt3) + " snitt4: " + tScoreD4(snitt4));
            sw.WriteLine("");
            sw.WriteLine(tScoreD4(snitt4));
			searcher.Close();
            sw.Close();
		}

        private double tScoreD3(ArrayList scores)
        {
            double numerator = 0.0;
            double denominator = 0.0;
            for (int j = 1; j <= scores.Count; ++j)
            {
                numerator += (((double)scores[j - 1]) * (1.0 / (double)j));
                denominator += (1.0 / (double)j);
            }
            return numerator / denominator;
        }

        private double tScoreD4(ArrayList scores)
        {
            double tScoreTest = 0.0;
            foreach (double s in scores)
                tScoreTest += s;
            return tScoreTest / scores.Count;
        }

        private void searchStandard()
        {
            DateTime start = DateTime.Now;

            try
            {
                searcher = new IndexSearcher(this.pathIndex);
            }
            catch (IOException ex)
            {

                Console.WriteLine("The index doesn't exist or is damaged. Please rebuild the index.\r\n\r\nDetails:\r\n" + ex.Message);
                return;
            }

            foreach (RelevanceSet trel in trelsSet)
            {
                
                Query query = QueryParser.Parse(trel.getQuery(), "text", analyzer); // query MISSING

                Hits hits = searcher.Search(query);           

                for (int i = 0; i < Math.Min(10, hits.Length()); i++)
                {
                    Console.WriteLine(trel.getQuery()+" "+hits.Score(i));
                    Document doc = hits.Doc(i);
                    string path = doc.Get("path");
                    FileInfo fi = new FileInfo(path);
                    string text = "";

                    if (fi.Extension == ".pdf")
                        text = AdobeParser.Parse(path);
                    else
                        text = Parser.Parse(path);

                    for (int t = 10; t >= 2; --t)
                    {
                        double meanSum = 0;
                        double sigmaSum = 0;
                        int meanCount = 0;

                        //for (int q = 0; q < 50; ++q)
                        //{
                            double sumScore = 0;
                            ArrayList scores = new ArrayList();

                            for (int j = 0; j < 100; ++j)
                            {
                                int numOn = 0;
                                int numOff = 0;

                                foreach (string ont in trel.getSubOnTopics(t))
                                {
                                    MatchCollection matches = Regex.Matches(text, ont);
                                    numOn += matches.Count;
                                }
                                foreach (string off in trel.getSubOffTopics(t))
                                {
                                    MatchCollection matches = Regex.Matches(text, off);
                                    numOff += matches.Count;
                                }

                                double tScore_basic = numOn - (BETA * numOff);
                                sumScore += tScore_basic;
                                scores.Add(tScore_basic);
                            }
                            double tmpSum = 0;
                            for (int j = 0; j < scores.Count; ++j)
                            {
                                tmpSum += Math.Pow(((double)scores[j] - (sumScore / scores.Count)), 2.0);
                            }
                            double sigma = Math.Sqrt(tmpSum / scores.Count);
                            //Console.WriteLine(t + " topics gives mean,standard deviation: " + (sumScore / scores.Count) + "  " + sigma);

                            double numerator = 0.0;
                            double denominator = 0.0;
                            for (int j = 1; j <= scores.Count; ++j)
                            {
                                numerator += ((double)scores[j - 1]) * (1 / ((double)j));
                                denominator += (1 / (double)j);
                            }
                            double tScoreDq = numerator / denominator;
                            //Console.WriteLine("Total tScore for "+t+" topics " + tScoreDq);
                        /*
                            using (StreamWriter sw = new StreamWriter("TestFile.txt"))
                            {
                                // Add some text to the file.
                                sw.Write("This is the ");
                                sw.WriteLine("header for the file.");
                                sw.WriteLine("-------------------");
                                // Arbitrary objects can also be written to the file.
                                sw.Write("The date is: ");
                                sw.WriteLine(DateTime.Now);
                            }
                        */
                            Console.WriteLine(tScoreDq);

                            meanSum += (sumScore / scores.Count);
                            sigmaSum += sigma;
                            meanCount++;
                        //}
                        //Console.WriteLine("Total for " + t + " topics. Mean: " + (meanSum / meanCount) + "  " + (sigmaSum / meanCount));
                        //Console.WriteLine("");
                    }
                }
                if (hits.Length() == 0) Console.WriteLine("No hits");
                    /*
                else
                {
                    double numerator = 0.0;
                    double denominator = 0.0;
                    for (int i = 1; i <= scores.Count; ++i)
                    {
                        numerator += ((double)scores[i - 1]) * (1 / ((double)i));
                        denominator += (1 / (double)i);
                    }
                    double tScoreDq = numerator / denominator;
                    Console.WriteLine("Total tScore: " + tScoreDq);

                    scores.Clear();
                }
                     * */
                Console.WriteLine();
            }
            searcher.Close();
        }

        private void searchNonRandom()
        {
            DateTime start = DateTime.Now;

            try
            {
                searcher = new IndexSearcher(this.pathIndex);
            }
            catch (IOException ex)
            {

                Console.WriteLine("The index doesn't exist or is damaged. Please rebuild the index.\r\n\r\nDetails:\r\n" + ex.Message);
                return;
            }

            foreach (RelevanceSet trel in trelsSet)
            {

                Query query = QueryParser.Parse(trel.getQuery(), "text", analyzer); // query MISSING

                Hits hits = searcher.Search(query);
                Console.WriteLine("Kalkulerer " + Math.Min(10, hits.Length()) + " hits for " + trel.getQuery());


                if (hits.Length() == 0)
                {
                    Console.WriteLine("No hits");
                    Console.WriteLine();
                }
                else
                {
                    ArrayList scores = new ArrayList();

                    for (int i = 0; i < Math.Min(10, hits.Length()); i++)
                    {
                        Console.WriteLine(trel.getQuery() + " " + hits.Score(i));
                        Document doc = hits.Doc(i);
                        string path = doc.Get("path");
                        FileInfo fi = new FileInfo(path);
                        string text = "";

                        if (fi.Extension == ".pdf")
                            text = AdobeParser.Parse(path);
                        else
                            text = Parser.Parse(path);

                        int numOn = 0;
                        int numOff = 0;

                        foreach (string ont in trel.getOnTopics())
                        {
                            MatchCollection matches = Regex.Matches(text, ont);
                            numOn += matches.Count;
                        }
                        foreach (string off in trel.getOffTopics())
                        {
                            MatchCollection matches = Regex.Matches(text, off);
                            numOff += matches.Count;
                        }

                        double tScore_basic = numOn - (BETA * numOff);
                        scores.Add(tScore_basic);
                    }

                    double numerator = 0.0;
                    double denominator = 0.0;
                    for (int j = 1; j <= scores.Count; ++j)
                    {
                        numerator += ((double)scores[j - 1]) * (1 / ((double)j));
                        denominator += (1 / (double)j);
                    }
                    double tScoreDq = numerator / denominator;
                    //Console.WriteLine("Total tScore for "+t+" topics " + tScoreDq);
                    /*
                        using (StreamWriter sw = new StreamWriter("TestFile.txt"))
                        {
                            // Add some text to the file.
                            sw.Write("This is the ");
                            sw.WriteLine("header for the file.");
                            sw.WriteLine("-------------------");
                            // Arbitrary objects can also be written to the file.
                            sw.Write("The date is: ");
                            sw.WriteLine(DateTime.Now);
                        }
                    */
                    Console.WriteLine(tScoreDq);
                }
            }
            searcher.Close();
        }

		private int smallerOf(int first, int second)
		{
			return first < second ? first : second;
		}
	}
}
